from hashlib import sha1
from types import *
from random import randint
from Crypto.Cipher import Blowfish

###########################
# Helpers
###########################

def braid2hex(b):
    return str2hex(str(b))

def hex2braid(h):
    return Braid(str2ary(hex2str(h)))

def str2hex(s):
    return ''.join(['%02X'%ord(c) for c in s])

def hex2str(h):
    return ''.join([chr(int(h[2*i:2*i+2],16)) for i in range(len(h)/2)])

def str2ary(s):
    return [ord(c) for c in s]
    
def ary2str(a):
    return ''.join([chr(i) for i in a])


###########################
# Braid classes
###########################

class Braid:

    """
    Braid basic class
    """

    def __init__(self,braid=None):
        if type(braid) is ListType:
            self.size = len(braid)
            self.items = braid
        elif type(braid) is IntType:
            self.items = [i for i in range(braid)]
            self.size = braid
        
    def __len__(self):
        return len(self.items)
        
    def __getitem__(self, key):
        return self.items[key]
    
    def __setitem__(self, key, value):
        self.swap(key,value)

    def swap(self, _in, _out):
        idx = self.items.index(_out)
        self.items[idx] = self.items[_in]
        self.items[_in] = _out
    
    def __repr__(self):
        return self.items
        
    def __str__(self):
        return ary2str(self.items)
        
    def reverse(self):
        rev = [self.items.index(i) for i in range(self.size)]
        return Braid(rev)
       
    def combine(self, _braid):
        if len(_braid) != self.size:
            raise "Invalid size"
        return Braid([_braid[self.items[i]] for i in range(self.size)])

    
    def shuffle(self,offset=0,size=0):
        for j in range(randint(1024,4096)):
            if size==0:
                for i in range(offset,self.size):
                    idx1 = randint(offset,self.size-1)
                    self.swap(i,idx1)
            else:
                for i in range(offset,size):
                    idx1 = randint(0,size-1)
                    self.swap(i,idx1)

# a little helper
def createBraidFromRawKey(rawkey):
    return Braid(str2ary(rawkey)) 

class BraidKey:

    """
    Key container based on braid asymmetric crypto.
    """
    
    def __init__(self, K, client):
        self.K = K
        N = len(K)
        self.privkey = Braid(N)
        if client:
            self.privkey.shuffle(offset=N/2)
        else:
            self.privkey.shuffle(size=N/2)
        self.privrkey = self.privkey.reverse()
        self.pubkey = self.privkey.combine(self.K.combine(self.privrkey))
    
    def getPublicKey(self):
        return str(self.pubkey)
        
    def getPrivateKey(self):
        return str(self.privkey)

class ClientBraidKey(BraidKey):
    """
    Braidkey for clients
    """
    def __init__(self, K):
        BraidKey.__init__(self, K, True)
        
class ServerBraidKey(BraidKey):
    """
    Braidkey for server
    """
    def __init__(self, K):
        BraidKey.__init__(self, K, False)


class SharedSecret:
    """
    This class is used to compute shared secret
    """
    
    def __init__(self, K, key):
        self.K = K
        self.Pk = key
        self.charset = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-+:;*$/][()|'
        
    def compute(self, pubkey):
        self.secret = self.Pk.privkey.combine(pubkey.combine(self.Pk.privrkey))
        
    def makeKey(self, pubkey):
        self.compute(pubkey)
        chars = ''.join(self.charset[c%len(self.charset)] for c in self.secret)
        h = sha1()
        h.update(chars)
        return h.hexdigest()
